package BankServer;

import ServerConfig.*;
import BankApp.*;
import org.omg.CosNaming.*;
import org.omg.CORBA.*;
import java.util.Properties;
import java.rmi.*;
import java.io.*;
import java.nio.*;
import java.util.*;
import java.net.*;
import java.util.concurrent.*;
import java.nio.channels.*;
import java.nio.channels.OverlappingFileLockException;

public class BankImpl extends BankPOA {

	public static ServerConfig localConfig = null; // The config of this server

	public boolean running = true; // False to notify outter wrapper to quit

	public boolean startFailed = true; // True to notify outter wrapper to quit

	public static UDPServerSocket UDPServer = null; // The socket to listen request

	public static ServerSocket transferServer = null; // The socket to accept transfer

	private ORB orb = null;

	//public String branchHostNames[];// ={"chile", "uruguay", "argentina"};
	//public static String branchHostNames[];//={"lynen", "lwoff"};

	public static int myBranchNo = -1;

	public static String myHostName = null;

	public BranchManager branchManager = null;

	public BankImpl(ServerConfig sc) throws Exception {
		localConfig = sc;
		myBranchNo = Integer.parseInt(sc.branchID);
		branchManager = new BranchManager(sc);
		UDPServer = branchManager.backupServer;
		transferServer = branchManager.serverSocket;
	}

	public void finalize() throws Exception {
		branchManager.finalize();
	}

	public void setORB(ORB orb_val) {
		orb = orb_val;
		branchManager.orb = orb;
	}

	public int Open(String opID) //throws java.rmi.RemoteException
	{

		// Debug
		BankUtility.showDebugInfo(1, "Open operation, opID=" + opID);
		
		int result = -1;
		try {
			result = branchManager.Open(opID);
		}
		catch (Exception er) {
			System.out.println(er.getMessage());
		}

		// Debug
		BankUtility.showDebugInfo(1, "Open operation returned with " + result);
		
		return result;
	}

	public float Deposit(String opID, int acnt, float amt) //throws java.rmi.RemoteException		
	{

		// Debug
		BankUtility.showDebugInfo(1, "Deposit operation, opID=" + opID + ", acnt=" + acnt + ", amt=" + amt);
		
		float result = BankUtility.NegativeInfinite;
		try {
			result = branchManager.Deposit(opID, acnt, amt);
		}
		catch (Exception er) {
			System.out.println(er.getMessage());
		}
		finally {

			// Debug
			BankUtility.showDebugInfo(1, "Deposit operation returned with " + result);
			
			return result;
		}
	}

	public float Withdraw(String opID, int acnt, float amt) //throws java.rmi.RemoteException
	{

		// Debug
		BankUtility.showDebugInfo(1, "Withdraw operation, opID=" + opID + ", acnt=" + acnt + ", amt=" + amt);
		
		float result = BankUtility.NegativeInfinite;
		try {
			result = branchManager.Withdraw(opID, acnt, amt);
		}
		catch (Exception er) {
			System.out.println(er.getMessage());
		}
		finally {

			// Debug
			BankUtility.showDebugInfo(1, "Withdraw operation returned with " + result);
			
			return result;
		}
	}

	public float Balance(String opID, int acnt) //throws java.rmi.RemoteException
	{

		// Debug
		BankUtility.showDebugInfo(1, "Balance operation, opID=" + opID + ", acnt=" + acnt);
		
		float result = BankUtility.NegativeInfinite;
		try {
			result = branchManager.Balance(opID, acnt);
		}
		catch (Exception er) {
			System.out.println(er.getMessage());
		}
		finally {

			// Debug
			BankUtility.showDebugInfo(1, "Balance operation returned with " + result);
			
			return result;
		}
	}

	public float Transfer(String opID, int src, int dest, float amt) //throws java.rmi.RemoteException
	{

		// Debug
		BankUtility.showDebugInfo(1, "Transfer operation, opID=" + opID + ", src=" + src + ", dest=" + dest + ", amt=" + amt);
		
		float result = BankUtility.NegativeInfinite;
		try {
			result = branchManager.Transfer(opID, src, dest, amt);
		}
		catch (Exception er) {
			System.out.println(er.getMessage());
		}
		finally {

			// Debug
			BankUtility.showDebugInfo(1, "Transfer operation returned with " + result);
			
			return result;
		}
	}

} //end class

class BranchManager extends java.lang.Object {
	public static ORB orb = null;
	
	//these are all static member which means only one copy is maintained
	public static final int AccountSize = BankUtility.ACCOUNT_SIZE;

	protected static Random rand = null;

	protected static boolean flag = true;

	private static final int MAX_AVAILABLE = 1;

	///////////////////////////////////////////////////////////////////////////////
	//THIS IS ACTING AS ENUM IN PACKET
	//REQUEST IS SENT BY INITIATOR	
	//protected final static int PREPARED=1;
	//THIS CAN BE REPLIED FROM RECEIVER OR SENT AS SENDER TO CANCEL
	//protected final static int REJECTED=2;

	///////////////////////////////////////////////////////////////////////////////
	protected static Semaphore available = null;

	//these are non-static member, the constructor and finalize should treat them differently
	protected int branchID = -1;

	//////////////////////////////////////////////////
	//to huang ming: this is the TCP socket	
	public ServerSocket serverSocket = null;

	public static RandomAccessFile branchFile = null;//, branchReader;

	protected RandomAccessFile rpcFiler = null, threadFiler = null;

	//////////////////////////////////////////////////////
	//mar. 26

	//two threads for listening
	protected static PrimaryListener primaryListener = null;

	protected static BackupListener backupListener = null;

	public static int serverMode = -1;

	protected static String backupHostName = null; //must be initialized

	protected static String myHostName = null;

	protected static int myBranchIndex = -1;

	public static String branchHostNames[] = null;

	public static int branchPortNumbers[] = null;

	protected static LogManager log = null;

	//protected static UDPClientSocket backupClient, primaryClients[];
	//protected static UDPServerSocket backupServer, primaryServer;

	///////////////////////////////////////////////////////
	//to huangming:
	//this is the socket you want
	public static UDPClientSocket backupClient = null;

	public static UDPServerSocket backupServer = null;

	//////////////////////////////////////////////////
	//the TCP port number for transferring
	protected static int primaryTransferPortNumber = -1;

	///////////////////////////////////////////////
	//this is the portnumber in backup for transfering log
	protected static int backupUDPPortNumber = -1;

	///////////////////////////////////////////////////////////////////////////
	//this is the portnumber for listening to request from backup
	protected static int localUDPPortNumber = -1;

	//////////////////////////////////////////////////////////////////////////////

	//protected static int transientCounter=0;

	protected static PacketWrapper packetWrapper = new PacketWrapper();

	protected static Packet packet = new Packet();

	protected static Semaphore modeSemaphore = null;

	protected static String myDataFileName = null, myLogFileName = null;

	protected boolean threadStop = false;

	///////////////////////////////////////////////////////////////////////////////////////////

	protected ServerConfig localConfig = null;

	public BranchManager(ServerConfig sc) throws Exception {
		localConfig = sc;
		branchID = Integer.parseInt(sc.branchID);

		//String fileName=Integer.toString(branchID)+".data";

		initialize(); //will call update and initialize all data

		backupServer = new UDPServerSocket(BranchManager.localUDPPortNumber);
		serverSocket = new ServerSocket(BranchManager.primaryTransferPortNumber);

		try {
			//non-static member
			rpcFiler = new RandomAccessFile(myDataFileName, "rw");
			threadFiler = new RandomAccessFile(myDataFileName, "rw");

			//backupSocket=new UDPClientSocket();
			//the following are all static global variables
			if (flag) {
				flag = false;
				//log=new LogManager(Integer.toString(branchID));
				log = new LogManager(myLogFileName);

				Date date = new Date();
				rand = new Random(date.getTime());
				available = new Semaphore(MAX_AVAILABLE, true);

				branchFile = new RandomAccessFile(myDataFileName, "rw");
				//branchReader=new RandomAccessFile(fileName, "r");
				if (branchFile.length() == 0)//this is the new file 
				{
					branchFile.setLength(AccountSize * 10000);
					branchFile.seek(0);//starting 
					for (int j = 0; j < 10000; j++) {
						//branchFiles[i].writeInt(0);//initialize to 0;
						branchFile.writeFloat(BankUtility.NegativeInfinite);// indicating unused
					}
				}

				//////////////////////////////////////////////////////////////////////////////
				//listener might need use "serversocket", so place it here

				//this method should be implemented to initialize all group information
				//!!!!!!!!!!this must be called before listening thread because 
				//portnumber need to be acquired

				////////////////////////
				/*
				 primaryClients=new UDPClientSocket[BankUtility.MaxBranchNumber];
				 for (int i=0; i<BankUtility.MaxBranchNumber; i++)
				 {
				 primaryClients[i]=new UDPClientSocket();
				 }
				 */
				modeSemaphore = new Semaphore(MAX_AVAILABLE, true);
				primaryListener = new PrimaryListener();
				backupListener = new BackupListener();

				primaryListener.start();
				backupListener.start();
				///////////////////////////////////////////////////////////			
			}
		}
		catch (Exception er) {
			System.out.println("BranchManager constructor error of " + er.getMessage());
			throw (er);
		}
	}

	////////////////////////////////////////////////////////////////////////////////////
	/////!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	//this method should be called every time there is a timeout
	//////////////!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	public synchronized void updateInformation() {
		ServerConfig sc = localConfig;
		String serverList[] = new String[0];
		try {
			serverList = ServerConfig.listServers(sc.rootContext);
		}
		catch (Exception ex) {
		}

		for (int i = 0; i < BankUtility.MaxBranchNumber; i++) {
			branchHostNames[i] = null;
			branchPortNumbers[i] = -1;
		}
		// Load those servers one by one
		for (int i = 0; i < serverList.length; i++) {
			// Retrieve current branchID and serverID
			String branchID = serverList[i].substring(6, 9);
			String serverID = serverList[i].substring(16, serverList[i].length());

			// Update the server list
			try {
				// Get the server configure
				ServerConfig server = ServerConfig.getServerConfig(sc.rootContext, branchID, true);
				branchHostNames[Integer.parseInt(branchID)] = server.localhostIP;
				branchPortNumbers[Integer.parseInt(branchID)] = Integer.parseInt(server.transferPort);
			}
			catch (Exception e) {
			}
		}

		try {
			this.primaryTransferPortNumber = Integer.parseInt(sc.transferPort);
			this.localUDPPortNumber = Integer.parseInt(sc.UDPPort);
			myDataFileName = sc.dataFileName;
			myLogFileName = sc.logFileName;

			String otherServerID = localConfig.serverID.compareTo("1")==0?"2":"1";
			ServerConfig backupConfig = ServerConfig.getServerConfig(sc.rootContext, sc.branchID, otherServerID);
			this.backupUDPPortNumber = Integer.parseInt(backupConfig.backupUDPPort);
			this.backupHostName = backupConfig.localhostIP;

		}
		catch (Exception e) {

		}

		/*
		 String theBranchID;
		 for (int i=0; i<BankUtility.MaxBranchNumber; i++)
		 {
		 try
		 {
		 theBranchID="branch"+100+i;
		 ServerConfig result=sc.getServerConfig(sc.rootContext, theBranchID, true);

		 branchHostNames[i]=InetAddress.getAllByName(result.localhostIP).getHostName();
		 branchPortNumbers[i]=Integer.parseInt(result.transferPort);
		 if (myBranchIndex==i)
		 {
		 primaryPortNumber=result.Integer.parseInt(result.transferPort);
		 backupPortNumber=result.UDPPort;
		 myDataFileName=result.dataFileName;
		 myLogFileName=result.logFileName;
		 result=sc.getServerConfig(sc.rootContext, theBranchID, false);
		 backupTransferPort=result.UDPPort;
		 backupHostName=InetAddress.getAllByName(result.localhostIP).getHostName();
		 }

		 }
		 catch (Exception er)
		 {
		 System.out.println("branch "+ branchID +" is not accessible");
		 }			
		 }
		 */
	}

	///////////////////////////////////////////////////////////////////////////////

	protected void initialize() {
		try {
			branchHostNames = new String[BankUtility.MaxBranchNumber];
			branchPortNumbers = new int[BankUtility.MaxBranchNumber];

			myBranchIndex = Integer.parseInt(localConfig.branchID);

			myHostName = localConfig.localhostIP;

			//ALWAYS STARTS WITH THIS MODE
			serverMode = BankUtility.SINGLE;
			// Debug
			BankUtility.showDebugInfo(1, "Server running in SINGLE mode.");

			//now call get server config
			updateInformation();

		}
		catch (Exception er) {
			System.out.println(er.getMessage());
		}

	}

	///////////////////////////////////////////////////////////////////////////////////////

	protected void finalize() throws java.rmi.RemoteException {
		try {
			if (serverSocket != null)
				serverSocket.close();//maybe it is better to stop thread then close socket.
			if (backupServer != null)
				backupServer.close();//maybe it is better to stop thread then close socket.
			if (branchFile != null) {
				branchFile.close();
				branchFile = null;
			}
			threadStop = true;
			if (rpcFiler != null)
				rpcFiler.close();
			if (threadFiler != null)
				threadFiler.close();

			////////////////////////////////////////////
			primaryListener.join();
			backupListener.join();
			/////////////////////////////////////////////
		}
		catch (Exception er) {
			throw new BankException("finalize error of " + er.getMessage());
		}
	}

	//////////////////////////////////////////////////////////////////////
	//this will be called by every method
	void modeCheck() {
		//////////////////////////////////////////////////
		try {
			while (serverMode == BankUtility.BACKUPJOIN) {
				Thread.sleep(BankUtility.WaitingLength);
			}
		}
		catch (Exception er) {
			System.out.println(er.getMessage());
		}
	}

	synchronized void startNewBackup() {
		try {
			// kh2, should not set the new check point, check point should be set only when the process totally succeeded
			// log.setCheckPoint();//set up a check point
			
			
			//change to server mode
			//////////////////////////////////////////////////////////////////////////////////////
			// to do 
			/////////////////////////////////////////////////////////////////////////////////
			//notify monitor	
			modeSemaphore.acquire();
			serverMode = BankUtility.SINGLE;
			// Debug
			BankUtility.showDebugInfo(1, "Can not contact with backup server, server changed to SINGLE mode.");
			modeSemaphore.release();
			//to do
			//call monitor to start a new backup and then

			// Notify monitor to load backup server
			ServerConfig.notifyMonitor(localConfig.rootContext, localConfig.branchID, false);

			updateInformation();//if necessary
		}
		catch (Exception er) {
			System.out.println(er.getMessage());
		}
	}

	synchronized void backupPacket(Packet packet) {
		//this is necessary if server is between state transfer	
		try {
			modeCheck();

			modeSemaphore.acquire();

			log.writePacket(packet);

			//we only send backup log when it is in normal state

			if (serverMode == BankUtility.NORMAL) {
				try {
					// Debug
					BankUtility.showDebugInfo(2, backupHostName + ":" + this.backupUDPPortNumber, packet);

					if(backupClient==null){
						this.backupClient = new UDPClientSocket();  
					}
					backupClient.sendPacket(backupHostName, this.backupUDPPortNumber, packet);
				}
				catch (Exception er) {
					if (er instanceof SocketException) {
						// Sending log packet failed, set the serverMode and relaunch a backup server
						modeSemaphore.release();
						startNewBackup();//should change mode, notify monitor that backup is down etc.
						modeSemaphore.acquire();
					}
					else {
						System.out.println("unknown error of backup packet" + er.getMessage());
					}
				}
			}

		}
		catch (Exception er) {
			System.out.println(er.getMessage());
		}
		finally{
			modeSemaphore.release();
		}

		//we must notify if we are ready for state transfer
		//THE REASON i DEFINE AN EXTRA STATE IS FOR SYNCHRONIZATION BETWEEN DIFFERENT THREAD
		//BECAUSE it is possible that when backup send join request while some of our transaction is still 
		//going on

	}

	////////////////////////////////////////////////////////////////////////

	//there is still need for file lock on opening new account
	//because you are trying to find invalid account
	//it is crazy to assume that user might use invalid account for transaction
	public synchronized int Open(String opID) throws java.rmi.RemoteException {
		/////////////////////////////////////////////////////////////////////////
		//we must wait for backup
		Packet packet;
		modeCheck();

		if (serverMode == BankUtility.SINGLE) {
			packet = log.findRecord(opID);
			if (packet != null) {
				//should be the source account
				return packet.source;
			}
		}

		/////////////////////////////////////////////////////////////////////////
		int account = BankUtility.InvalidAccountNumber;
		boolean result = false;
		FileLock fileLock = null;
		try {
			while (!result) {
				//allow account starting from 0
				account = rand.nextInt(10000);
				if ((fileLock = lockFile(account, false)) != null) {
					//System.out.println("get semaphore");
					fileLock.release();
					rpcFiler.seek(account * AccountSize);
					// Check if this account is opened
					if (rpcFiler.readFloat() == BankUtility.NegativeInfinite) {
						// Write the log in local file and backup it on remote backup server
						packet = new Packet();
						packet.opID = opID;
						packet.opType = Packet.OPEN;
						packet.source = Integer.parseInt(localConfig.branchID)*10000+account;
						packet.dest = -1;//ignored
						packet.oldValue = BankUtility.NegativeInfinite;
						packet.newValue = 0;//is initialized to 0
						packet.tag = -1;//should be ignored
						packet.status=BankUtility.SUCCESS;//Indicate the succeeded of writting log
						
						backupPacket(packet);

						// Write to local file
						rpcFiler.seek(account * AccountSize);
						rpcFiler.writeFloat(0);
						result = true;
						
						// Operation succeeded, set checkpoint
						if(serverMode==BankUtility.NORMAL) log.setCheckPoint();
						
						account += Integer.parseInt(localConfig.branchID)*10000;
					}
					//always release! and don't use "break" which is bad!!!!
					available.release();
				}
			}
		}
		catch (Exception er) {
			System.out.println(er.getMessage());	
			account = BankUtility.InvalidAccountNumber;//indicating unknown error
			throw new BankException("Error in open operation " + er.getMessage());
		}
		finally {
			return account;
		}
	}

	//we first try to get file lock with possible several failure
	//then we get semaphore and r/w record
	//finally must test file lock and release it, we cannot do this for semaphore
	//so, whenever exception happens, it must maintain a correct counter for it.
	public synchronized float Deposit(String opID, int acnt, float amount) throws java.rmi.RemoteException {
		/////////////////////////////////////////////////////////////////////
		modeCheck();
		if (serverMode == BankUtility.SINGLE) {
			packet = log.findRecord(opID);
			if (packet != null) {
				//should be the source account
				return packet.newValue;
			}
		}

		////////////////////////////////////////////////////////////////////

		FileLock fileLock = null;
		int account = acnt % 10000;
		int counter = 0;

		float balance = BankUtility.NegativeInfinite;
		//entering critical section
		try {
			//account is possibly locked so we need to try several times to lock
			if ((fileLock = lockFile(account, false)) != null) {
				fileLock.release();
				rpcFiler.seek(account * AccountSize);
				balance = rpcFiler.readFloat();
				// Check if account is opened
				if (balance != BankUtility.NegativeInfinite) {
					balance += amount;
					
					// Write the log in local file and backup it on remote backup server
					Packet packet = new Packet();
					packet.opID = opID;
					//packet.hostIndex=hostIndex;
					packet.opType = Packet.DEPOSIT;
					packet.source = Integer.parseInt(localConfig.branchID)*10000+account;
					packet.dest = -1;//ignored
					packet.oldValue = balance - amount;
					packet.newValue = balance;//is initialized to 0
					packet.tag = -1;//should be ignored
					packet.status=BankUtility.SUCCESS;//Indicate the succeeded of writting log
					backupPacket(packet);

					// Write to local file
					rpcFiler.seek(account * AccountSize);
					rpcFiler.writeFloat(balance);

					// Operation succeeded, set checkpoint
					if(serverMode==BankUtility.NORMAL) log.setCheckPoint();
				}
				available.release();
			}

		}
		catch (Exception er) {
			balance = BankUtility.NegativeInfinite;
			throw new BankException("unknown deposit error: " + er.getMessage());
		}
		finally {
			return balance;
		}
	}

	//assume amount is positive
	//the sequence of filelock and semaphore is critical in performance
	//we only need to lock and want to hold semaphore as short as possible
	public synchronized float Withdraw(String opID, int acnt, float amount) throws java.rmi.RemoteException {
		/////////////////////////////////////////////////////////////////////
		modeCheck();
		if (serverMode == BankUtility.SINGLE) {
			packet = log.findRecord(opID);
			if (packet != null) {
				//should be the source account
				return packet.newValue;
			}
		}

		////////////////////////////////////////////////////////////////////
		FileLock fileLock = null;
		int account = acnt % 10000;
		int counter = 0;
		float balance = BankUtility.NegativeInfinite;
		try {
			//account is possibly locked so we need to try several times to lock
			if ((fileLock = lockFile(account, false)) != null) {
				fileLock.release();
				rpcFiler.seek(account * AccountSize);
				balance = rpcFiler.readFloat();
				// Check if account is opened
				if (balance != BankUtility.NegativeInfinite && balance >= amount) {
					balance -= amount;

					// Write the log in local log file and backup it on remote backup server
					Packet packet = new Packet();
					packet.opID = opID;
					//packet.hostIndex=hostIndex;
					packet.opType = Packet.WITHDRAW;
					packet.source = Integer.parseInt(localConfig.branchID)*10000+account;
					packet.dest = -1;//ignored
					packet.oldValue = balance + amount;
					packet.newValue = balance;//is initialized to 0
					packet.tag = -1;//should be ignored
					packet.status=BankUtility.SUCCESS;//Indicate the succeeded of writting log
					backupPacket(packet);

					// Write to local log file
					rpcFiler.seek(account * AccountSize);
					rpcFiler.writeFloat(balance);

					// Operation succeeded, set checkpoint
					if(serverMode==BankUtility.NORMAL) log.setCheckPoint();
				}
				else {
					balance = BankUtility.NegativeInfinite;
				}
				available.release();
			}

		}
		catch (Exception er) {
			balance = BankUtility.NegativeInfinite;
			throw new BankException("unknown withdraw error: " + er.getMessage());
		}
		finally {
			return balance;
		}
	}

	//even reading may not need to be synchronized, but 
	//file pointer has to be! So make it simple by synchronized!
	public synchronized float Balance(String opID, int acnt) throws java.rmi.RemoteException {
		/////////////////////////////////////////////////////////////////////
		//there is no record to check but we still need to hold "file-reading"
		//modeCheck();

		//we don't need hold this, because the log file will not be touched by this method

		////////////////////////////////////////////////////////////////////

		float balance = BankUtility.NegativeInfinite;
		FileLock fileLock = null;
		int account = acnt % 10000;
		int counter = 0;
		try {
			//account is possibly locked so we need to try several times to lock
			if ((fileLock = lockFile(account, true)) != null) {
				fileLock.release();
				rpcFiler.seek(account * AccountSize);
				balance = rpcFiler.readFloat();
				//account not opened
				available.release();
			}
		}
		catch (Exception er) {
			balance = BankUtility.NegativeInfinite;
			throw new BankException("unknown balance error: " + er.getMessage());
		}
		finally {
			return balance;
		}
	}

	//
	protected synchronized float localTransfer(String opID, int src, int dest, float amt) throws java.rmi.RemoteException {

		FileLock srcLock = null, destLock = null;
		float srcValue = BankUtility.NegativeInfinite, destValue = BankUtility.NegativeInfinite;
		int srcAccount = src % 10000, destAccount = dest % 10000;
		try {
			//notice the style, we have to do like this
			if ((srcLock = lockFile(srcAccount, false)) != null) {
				available.release();
				if ((destLock = lockFile(destAccount, false)) != null) {
					srcLock.release();
					destLock.release();
					rpcFiler.seek(srcAccount * AccountSize);
					srcValue = rpcFiler.readFloat();
					rpcFiler.seek(destAccount * AccountSize);
					destValue = rpcFiler.readFloat();

					if (srcValue != BankUtility.NegativeInfinite && srcValue >= amt) {
						srcValue -= amt;
						destValue += amt;
						
						// Write the log in local log file and backup it on remote backup server
						Packet packet = new Packet();
						packet.opID = opID;
						packet.opType = BankUtility.TRANSFER;
						packet.source = src;
						packet.dest = dest;
						packet.oldValue = srcValue + amt;
						packet.newValue = srcValue;//is initialized to 0
						packet.tag = -1;//should be ignored
						packet.status=BankUtility.SUCCESS;//should be ignored
						backupPacket(packet);

						// Write to local data file
						rpcFiler.seek(srcAccount * AccountSize);
						rpcFiler.writeFloat(srcValue);
						rpcFiler.seek(destAccount * AccountSize);
						rpcFiler.writeFloat(destValue);

						// Operation succeeded, set checkpoint
						if(serverMode==BankUtility.NORMAL) log.setCheckPoint();
					}
					else {
						srcValue = BankUtility.NegativeInfinite;
					}
					available.release();
				}
				else {
					srcLock.release();
				}
			}
		}
		catch (Exception er) {
			throw new BankException("unknown error of local transfer:" + er.getMessage());
		}
		finally {
			return srcValue;
		}
	}

	protected synchronized float remoteTransfer(String opID, int src, int dest, float amount) throws java.rmi.RemoteException {
		//first we send out our transfer request
		float srcValue = BankUtility.NegativeInfinite;
		int account = src % 10000;
		//int reply, index=dest/10000-100;
		int reply, index = dest / 10000;
		FileLock fileLock = null;

		try {
			//checkHostName(src);//possible has exception

			if ((fileLock = lockFile(account, false)) != null)//do nothing
			{
				//System.out.println("after remote transfer for file lock");
				fileLock.release();
				rpcFiler.seek(account * AccountSize);
				srcValue = rpcFiler.readFloat();
				if (srcValue != BankUtility.NegativeInfinite && srcValue >= amount) {
					//here we lock again because rpc is slow, we give other one chances to lock
					//System.out.println("before re-lock in remote transfer for file lock");
					try {
						fileLock = branchFile.getChannel().lock(account * AccountSize, AccountSize, false);
					}
					catch (Exception er) {
						System.out.println("This cannot be true, if it happens everything is wrong");
					}
					available.release();

					//////////////////////////////////////////////////////////////////////////////////////////
					/*
					 Socket socket=new Socket(BankServer.branchHostNames[index], BankUtility.BasePortNumber);

					 java.io.DataOutputStream  outStream=new java.io.DataOutputStream(socket.getOutputStream());
					 java.io.DataInputStream  inStream=new java.io.DataInputStream(socket.getInputStream());					
					 
					 outStream.writeInt(dest);//account
					 outStream.writeFloat(amount);//amount
					 
					 reply=inStream.readInt();				
					 //must close socket, file resource, in java it is impossible to do this for you
					 
					 */

					boolean done;

					srcValue -= amount;
					
					// Write the log in local log file and backup it on remote backup server
					Packet packet = new Packet();
					packet.opID = opID;
					packet.opType = BankUtility.TRANSFER;
					packet.source = src;
					packet.dest = dest;
					packet.oldValue = srcValue + amount;
					packet.newValue = srcValue;//is initialized to 0
					packet.tag = -1;//should be ignored
					packet.status=BankUtility.SUCCESS;//should be ignored
					backupPacket(packet);


					// Setup the packet for transfer operation
					Packet transferPacket = new Packet();
					transferPacket.opID=opID;
					transferPacket.opType=BankUtility.TRANSFER;
					transferPacket.source=src;
					transferPacket.dest=dest;
					transferPacket.oldValue=amount;//this is no good;
					transferPacket.newValue=-1;//should be ignored
					
					
					done = true;
					try {
						/*
						if (branchHostNames[index] == null) {
							updateInformation();
							if (branchHostNames[index] == null)
								throw new Exception();
						}
						*/
						updateInformation();
						if (branchHostNames[index] == null)
							throw new Exception();
						
						// Debug
						BankUtility.showDebugInfo(1, "Using TCP to connect to " + branchHostNames[index] + ":" + branchPortNumbers[index]);
						
						Socket socket = new Socket(branchHostNames[index], branchPortNumbers[index]);

						packetWrapper.writePacket(socket, transferPacket);
						packet = packetWrapper.readPacket(socket);

					}
					catch (Exception er) {
						done = false;
						packet.status = BankUtility.TRANSFER_REJECTED;
					}

					reply = packet.status;

					//////////////////////////////////////////////////////////////////////////////////////////

					if (reply == BankUtility.TRANSFER_PREPARED) {
						//WE ARE SURE WE ARE THE SOLE USER OF RPC. AND WE ARE RPC USER NOW!!!!!, NOT THREAD USER
						available.acquire();//we need to prevent others accidentaly lock again
						fileLock.release();

						// Write local data file
						rpcFiler.seek(account * AccountSize);
						rpcFiler.writeFloat(srcValue);
						available.release();

						// Operation succeeded, set checkpoint
						if(serverMode==BankUtility.NORMAL) log.setCheckPoint();
					}
					else {
						fileLock.release();
						srcValue = BankUtility.NegativeInfinite;
						//we didn't acquire semaphore
						if (reply != BankUtility.TRANSFER_REJECTED) {
							throw new BankException("Communication with remote branch server failed.");
						}
						//we still need to release, but we don't have to release semaphore

					}
				}
				else {
					srcValue = BankUtility.NegativeInfinite;//this costs me 5 hours!!!!!!!
					available.release();
				}
			}
		}
		catch (Exception er) {

			srcValue = BankUtility.NegativeInfinite;
			throw new BankException("Error of remote transfer:" + er.getMessage());
		}
		finally {
			return srcValue;
		}
	}

	//we use a dummy global filepointer branchFile for lock
	//and use another file pointer for writing
	protected FileLock lockFile(int account, boolean toShare) {
		int counter = 0;
		FileLock fileLock = null;
		try {
			while (true) {
				//allow account starting from 0						
				fileLock = null;
				available.acquire();
				try {
					fileLock = branchFile.getChannel().tryLock(account * AccountSize, AccountSize, toShare);
				}
				catch (Exception er) {
					if (!(er instanceof OverlappingFileLockException)) {
						System.out.println("what?");
					}
				}
				//critical section should be as short as possible
				//available.release();//must release to give other a chance to try

				if (fileLock == null) {
					available.release();
					if (counter < BankUtility.MaxLockTrialNumber) {
						counter++;
						Thread.sleep(BankUtility.SleepingLength);
						//notify();
						//wait();
					}
					else {
						//cannot acquire lock, now quit						
						//return oldValue;//failed						
						throw new BankException(" account is locked");
					}
				}
				else {
					//don't release semaphore
					break;
				}
			}
		}
		catch (Exception er) {
			fileLock = null;//indicating failure
			//System.out.println("lock file error of "+er.getMessage());
		}
		finally {
			return fileLock;
		}
	}

	public synchronized float Transfer(String opID, int src, int dest, float amt) throws java.rmi.RemoteException {
		/////////////////////////////////////////////////////////////////////
		modeCheck();
		if (serverMode == BankUtility.SINGLE) {
			packet = log.findRecord(opID);
			if (packet != null) {
				//should be the source account
				return packet.newValue;
			}
		}

		////////////////////////////////////////////////////////////////////

		float result = BankUtility.NegativeInfinite;
		try {
			if (dest / 10000 == this.branchID) {
				//do the local transfer without using TCP
				result = localTransfer(opID, src, dest, amt);
			}
			else {
				result = remoteTransfer(opID, src, dest, amt);
			}
		}
		catch (Exception er) {
			result = BankUtility.NegativeInfinite;
			throw new BankException("unknown error of transfer:" + er.getMessage());
		}
		finally {
			return result;
		}
	}

	class BackupListener extends Thread {
		public void run() {
			UDPClientSocket sender = new UDPClientSocket();
			int selfChecking = 0;
			while (!threadStop) {
				try {
					Packet result = backupServer.recvPacket();
					
					if (result.status == BankUtility.REQUEST_JOIN) {
						
						ServerConfig backupConfig = null;
						try {
							String otherServerID = localConfig.serverID.compareTo("1")==0?"2":"1";
							backupConfig = ServerConfig.getServerConfig(localConfig.rootContext, localConfig.branchID,otherServerID);
						}
						catch (Exception e) {
						}
						backupUDPPortNumber = Integer.parseInt(backupConfig.backupUDPPort);
						backupHostName = backupConfig.localhostIP;

//						if (serverMode == BankUtility.SINGLE) {
							Packet packet;
							try {
								// UDPClientSocket sender=new UDPClientSocket();
								modeSemaphore.acquire();
								serverMode = BankUtility.BACKUPJOIN;
								// Debug
								BankUtility.showDebugInfo(1, "Received join request from backup server, server changed to BACKUPJOIN mode.");
								log.gotoCheckPoint();
								boolean done = false;
								do {
									// Read next log
									try {
										packet = log.readPacket();
									}
									catch (Exception e) {
										// Reach end of log file, join process finished
										done = true;
										continue;
									}
									
									// No packet need to send
									if(packet == null){
										done=true;
										continue;
									}
									
									// Send to backup server
									try{
										sender.sendPacket(backupHostName, backupUDPPortNumber, packet);
									}catch(Exception e){
										// Sending to backup server failed, resume to single status
										done = true;
										serverMode = BankUtility.SINGLE;

										// Debug
										BankUtility.showDebugInfo(1, "Backup server failed during join process, server changed to SINGLE mode.");

										throw new Exception("Backup server failed during join process.");
									}
									
									// Operation succeeded, move forward the check point
									log.setCheckPoint(log.getCurrentPos());
								} while (!done);

								// Debug
								BankUtility.showDebugInfo(1, "Backup server join process finished, server changed to NORMAL mode.");

								serverMode = BankUtility.NORMAL;

							}
							catch (Exception er) {
								// Error while joining process
								// System.out.println("error in backup listener " + er.getMessage());
							}
							finally{
								modeSemaphore.release();
							}
//						}
					}
				}
				catch (Exception e) {
					// System.out.println("UDP listener error of "+e.getMessage());
				}

				/* Depreciated				
				 if (result.status==BankUtility.BACKUP_READY)
				 {
				 if (serverMode==BankUtility.BACKUPJOIN)
				 {
				 try
				 {
				 modeSemaphore.acquire();
				 serverMode=BankUtility.NORMAL;
				 modeSemaphore.release();
				 }
				 catch (Exception er)
				 {
				 System.out.println("error in backup listener when try to rewrite mode"+er.getMessage());
				 }
				 }
				 }
				 */
			} // End of while
			
			// Clean up
			if (backupServer != null)
				backupServer.close();
			if (sender != null)
				sender.close();
		}
	}

	//this 
	class PrimaryListener extends Thread {
		public void run() {
			FileLock fileLock = null;
			int result, account;
			int selfChecking = 0;
			float amount = -1, oldValue;
			while (!threadStop) {
				try {
					serverSocket.setSoTimeout(BankUtility.WaitingLength);
					Socket socket = serverSocket.accept();

					//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					packet = packetWrapper.readPacket(socket);
					amount = packet.oldValue;//this is just a convention
					account = packet.dest % 10000;//the 

					// Debug
					BankUtility.showDebugInfo(1, "Received transfer request, opID=" + packet.opID + ", from " + packet.source + " to " + packet.dest + " amount is " + amount);

					//if (packet.status == BankUtility.TRANSFER_CHECK) {
					
					// Check if this transfer operation has been processed
						modeSemaphore.acquire();
						Packet oldPacket = log.findRecord(packet.opID);
						if (oldPacket != null) {
							packet.status = BankUtility.TRANSFER_PREPARED;
							packetWrapper.writePacket(socket, packet);
							modeSemaphore.release();
							continue;
						}
						modeSemaphore.release();
					//}

					////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

					/*
					 inStream=new java.io.DataInputStream(socket.getInputStream());
					 outStream=new java.io.DataOutputStream(socket.getOutputStream());


					 account=inStream.readInt()%10000;//the index of source branch										
					 amount=inStream.readFloat();
					 */

					////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					if ((fileLock = lockFile(account, false)) != null) {
						//System.out.println("after listener trying to lock file");
						fileLock.release();
						threadFiler.seek(account * AccountSize);
						oldValue = threadFiler.readFloat();
						if (oldValue != BankUtility.NegativeInfinite) {
							oldValue += amount;
							
							// Write the log in local log file and backup it on remote backup server
							packet.status = BankUtility.SUCCESS;
							backupPacket(packet);
							
							// Write to local data file
							threadFiler.seek(account * AccountSize);
							threadFiler.writeFloat(oldValue);
							result = BankUtility.TRANSFER_PREPARED;

							// Operation succeeded, set checkpoint
							if(serverMode==BankUtility.NORMAL) log.setCheckPoint();
						}
						else {
							result = BankUtility.TRANSFER_REJECTED;
						}
						available.release();
					}
					else {
						result = BankUtility.TRANSFER_REJECTED;
					}

					///////////////////////////////////////////////////////////////////////////////////////////////////
					packet.status = result;
					packetWrapper.writePacket(socket, packet);

					if (packet.status == BankUtility.TRANSFER_PREPARED) {
						// backupPacket(packet);
					}

					///////////////////////////////////////////////////////////////////////////////////////////////////////
					/*
					 outStream.writeInt(result);								
					 */
					///////////////////////////////////////////////////////////////////////////////////////////////////////
				}
				catch (Exception er) {
					//System.out.println("listener error of "+er.getMessage());
				}
			}
		}
	}
}
